from typing import Optional
import yaml
import json
import os
from ._misc import sentencecase_to_pascalcase
from .llm_connect import GptConversation


class BioCypherPromptEngine:
    def __init__(
        self,
        schema_config_or_info_path: Optional[str] = None,
        schema_config_or_info_dict: Optional[dict] = None,
    ):
        """

        Given a biocypher schema configuration, extract the entities and
        relationships, and for each extract their mode of representation (node
        or edge), properties, and identifier namespace. Using these data, allow
        the generation of prompts for a large language model, informing it of
        the schema constituents and their properties, to enable the
        parameterisation of function calls to a knowledge graph.

        Args:
            schema_config_or_info_path: Path to a biocypher schema configuration
                file or the extended schema information output generated by
                BioCypher's `write_schema_info` function (preferred).

            schema_config_or_info_dict: A dictionary containing the schema
                configuration file or the extended schema information output
                generated by BioCypher's `write_schema_info` function
                (preferred).

        """

        if not schema_config_or_info_path and not schema_config_or_info_dict:
            raise ValueError(
                "Please provide the schema configuration or schema info as a "
                "path to a file or as a dictionary."
            )

        if schema_config_or_info_path and schema_config_or_info_dict:
            raise ValueError(
                "Please provide the schema configuration or schema info as a "
                "path to a file or as a dictionary, not both."
            )

        if schema_config_or_info_path:
            # read the schema configuration
            with open(schema_config_or_info_path, "r") as f:
                schema_config = yaml.safe_load(f)
        elif schema_config_or_info_dict:
            schema_config = schema_config_or_info_dict

        # check whether it is the original schema config or the output of
        # biocypher info
        is_schema_info = schema_config.get("is_schema_info", False)

        # extract the entities and relationships: each top level key that has
        # a 'represented_as' key
        self.entities = {}
        self.relationships = {}
        if not is_schema_info:
            for key, value in schema_config.items():
                # hacky, better with biocypher output
                name_indicates_relationship = (
                    "interaction" in key.lower() or "association" in key.lower()
                )
                if "represented_as" in value:
                    if (
                        value["represented_as"] == "node"
                        and not name_indicates_relationship
                    ):
                        self.entities[sentencecase_to_pascalcase(key)] = value
                    elif (
                        value["represented_as"] == "node"
                        and name_indicates_relationship
                    ):
                        self.relationships[
                            sentencecase_to_pascalcase(key)
                        ] = value
                    elif value["represented_as"] == "edge":
                        self.relationships[
                            sentencecase_to_pascalcase(key)
                        ] = value
        else:
            for key, value in schema_config.items():
                if not isinstance(value, dict):
                    continue
                if value.get("present_in_knowledge_graph", None) == False:
                    continue
                if value.get("is_relationship", None) == False:
                    self.entities[sentencecase_to_pascalcase(key)] = value
                elif value.get("is_relationship", None) == True:
                    self.relationships[sentencecase_to_pascalcase(key)] = value

        self.question = ""
        self.selected_entities = []
        self.selected_relationships = []
        self.selected_relationship_labels = []  # copy to deal with labels that
        # are not the same as the relationship name

    def generate_query(self, question: str, query_language: str) -> str:
        """
        Wrap entity and property selection and query generation; return the
        generated query.

        Args:
            question: A user's question.

            query_language: The language of the query to generate.

        Returns:
            A database query that could answer the user's question.
        """

        success1 = self._select_entities(question)
        if not success1:
            raise ValueError(
                "Entity selection failed. Please try again with a different "
                "question."
            )
        success2 = self._select_properties()
        if not success2:
            raise ValueError(
                "Property selection failed. Please try again with a different "
                "question."
            )

        return self._generate_query(
            question=question,
            entities=self.selected_entities,
            relationships=self.selected_relationship_labels,
            properties=self.selected_properties,
            query_language=query_language,
        )

    def _select_entities(self, question: str) -> bool:
        """

        Given a question, select the entities that are relevant to the question
        and store them in `selected_entities` and `selected_relationships`. Use
        LLM conversation to do this.

        Args:
            question: A user's question.

        Returns:
            True if at least one entity was selected, False otherwise.

        """

        self.question = question

        conversation = GptConversation(
            model_name="gpt-3.5-turbo",
            prompts={},
            correct=False,
        )

        conversation.set_api_key(
            api_key=os.getenv("OPENAI_API_KEY"), user="entity_selector"
        )

        conversation.append_system_message(
            (
                "You have access to a knowledge graph that contains "
                f"these entities: {', '.join(self.entities)} and these "
                f"relationships: {', '.join(self.relationships)}. Your task is "
                "to select the ones that are relevant to the user's question "
                "for subsequent use in a query. Only return the entities and "
                "relationships, comma-separated, without any additional text. "
                "If you select relationships, make sure to also return "
                "entities that are connected by those relationships."
            )
        )

        msg, token_usage, correction = conversation.query(question)

        result = msg.split(",") if msg else []
        # TODO: do we go back and retry if no entities were selected? or ask for
        # a reason? offer visual selection of entities and relationships by the
        # user?

        if result:
            for entity_or_relationship in result:
                entity_or_relationship = entity_or_relationship.strip()
                if entity_or_relationship in self.entities:
                    self.selected_entities.append(entity_or_relationship)
                elif entity_or_relationship in self.relationships:
                    self.selected_relationships.append(entity_or_relationship)
                    self.selected_relationship_labels.append(
                        self.relationships[entity_or_relationship].get(
                            "label_as_edge", entity_or_relationship
                        )
                    )

        return bool(result)

    def _select_properties(
        self,
        question: Optional[str] = None,
        entities: Optional[list] = None,
        relationships: Optional[list] = None,
    ):
        """

        Given a question (optionally provided, but in the standard use case
        reused from the entity selection step) and the selected entities, select
        the properties that are relevant to the question and store them in
        the dictionary `selected_properties`.

        Args:
            question (Optional[str]): A user's question.

        Returns:
            True if at least one property was selected, False otherwise.

        """

        question = question or self.question

        if not question:
            raise ValueError(
                "No question provided, and no question from entity selection "
                "step available. Please provide a question or run the "
                "entity selection (`select_entities()`) step first."
            )

        entities = entities or self.selected_entities
        relationships = relationships or self.selected_relationships

        if not entities and not relationships:
            raise ValueError(
                "No entities or relationships provided, and none available "
                "from entity selection step. Please provide "
                "entities/relationships or run the entity selection "
                "(`select_entities()`) step first."
            )

        e_props = {}
        for entity in entities:
            if self.entities[entity].get("properties"):
                e_props[entity] = list(
                    self.entities[entity]["properties"].keys()
                )

        r_props = {}
        for relationship in relationships:
            if self.relationships[relationship].get("properties"):
                r_props[relationship] = list(
                    self.relationships[relationship]["properties"].keys()
                )

        # TODO: split into separate prompts for entities and relationships,
        # return single JSON each

        msg = (
            "You have access to a knowledge graph that contains entities and "
            "relationships. They have the following properties. Entities:"
            f"{e_props}, Relationships: {r_props}. "
            "Your task is to select the properties that are relevant to the "
            "user's question for subsequent use in a query. Only return the "
            "entities and relationships with their relevant properties in JSON "
            "format, without any additional text. Return the "
            "entities/relationships as top-level dictionary keys, and their "
            "properties as dictionary values. "
            "Do not return properties that are not relevant to the question."
        )

        conversation = GptConversation(
            model_name="gpt-3.5-turbo",
            prompts={},
            correct=False,
        )

        conversation.set_api_key(
            api_key=os.getenv("OPENAI_API_KEY"), user="property_selector"
        )

        conversation.append_system_message(msg)

        msg, token_usage, correction = conversation.query(question)

        self.selected_properties = json.loads(msg) if msg else {}

        return bool(self.selected_properties)

    def _generate_query(
        self,
        question: str,
        entities: list,
        relationships: list,
        properties: dict,
        query_language: str,
    ):
        """
        Generate a query in the specified query language that answers the user's
        question.

        Args:
            question: A user's question.

            entities: A list of entities that are relevant to the question.

            relationships: A list of relationships that are relevant to the
                question.

            properties: A dictionary of properties that are relevant to the
                question.

            query_language: The language of the query to generate.

        TODO model needs indication of directionality of relationships
        """
        msg = (
            f"Generate a database query in {query_language} that answers "
            f"the user's question. "
            f"You can use the following entities: {entities}, "
            f"relationships: {relationships}, and properties: {properties}. "
            "Only return the query, without any additional text."
        )

        conversation = GptConversation(
            model_name="gpt-3.5-turbo",
            prompts={},
            correct=False,
        )

        conversation.set_api_key(
            api_key=os.getenv("OPENAI_API_KEY"), user="query_generator"
        )

        conversation.append_system_message(msg)

        msg, token_usage, correction = conversation.query(question)

        return msg



# ChatGSE document summarisation functionality
# split text
# connect to vector db
# do similarity search
# return x closes matches for in-context learning

from typing import List, Optional, Dict

from langchain.schema import Document
from langchain.embeddings.openai import OpenAIEmbeddings
from langchain.text_splitter import RecursiveCharacterTextSplitter
from langchain.document_loaders import TextLoader
from langchain.vectorstores import Milvus

import fitz  # this is PyMuPDF (PyPI pymupdf package, not fitz)
from transformers import GPT2TokenizerFast

from biochatter.vectorstore_host import VectorDatabaseHostMilvus


class DocumentEmbedder:
    def __init__(
        self,
        use_prompt: bool = True,
        used: bool = False,
        online: bool = False,
        chunk_size: int = 1000,
        chunk_overlap: int = 0,
        split_by_characters: bool = True,
        separators: Optional[list] = None,
        n_results: int = 3,
        model: Optional[str] = None,
        vector_db_vendor: Optional[str] = None,
        connection_args: Optional[dict] = None,
        api_key: Optional[str] = None,
    ) -> None:
        self.use_prompt = use_prompt
        self.used = used
        self.online = online
        self.chunk_size = chunk_size
        self.chunk_overlap = chunk_overlap
        self.separators = separators or [" ", ",", "\n"]
        self.n_results = n_results
        self.split_by_characters = split_by_characters
        self.model_name = model

        # TODO: variable embeddings depending on model
        # for now, just use ada-002
        # TODO API Key handling to central config
        if not self.online:
            self.embeddings = OpenAIEmbeddings(openai_api_key=api_key)
        else:
            self.embeddings = None

        # connection arguments
        self.connection_args = connection_args or {
            "host": "127.0.0.1",
            "port": "19530",
        }

        # TODO: vector db selection
        self.vector_db_vendor = vector_db_vendor or "milvus"
        # instantiate VectorDatabaseHost
        self.database_host = None
        self._init_database_host()

    def _init_database_host(self):
        if self.vector_db_vendor == "milvus":
            self.database_host = VectorDatabaseHostMilvus(
                embedding_func=self.embeddings, connection_args=self.connection_args
            )
        else:
            raise NotImplementedError(self.vector_db_vendor)

    def set_chunk_siue(self, chunk_size: int) -> None:
        self.chunk_size = chunk_size

    def set_chunk_overlap(self, chunk_overlap: int) -> None:
        self.chunk_overlap = chunk_overlap

    def set_separators(self, separators: list) -> None:
        self.separators = separators

    def _characters_splitter(self) -> RecursiveCharacterTextSplitter:
        return RecursiveCharacterTextSplitter(
            chunk_size=self.chunk_size,
            chunk_overlap=self.chunk_overlap,
            separators=self.separators,
        )

    def _tokens_splitter(self) -> RecursiveCharacterTextSplitter:
        DEFAULT_OPENAI_MODEL = "gpt-3.5-turbo"
        HUGGINGFACE_MODELS = ["bigscience/bloom"]
        if self.model_name and self.model_name in HUGGINGFACE_MODELS:
            tokenizer = GPT2TokenizerFast.from_pretrained("gpt2")
            return RecursiveCharacterTextSplitter.from_huggingface_tokenizer(
                tokenizer,
                chunk_size=self.chunk_size,
                chunk_overlap=self.chunk_overlap,
                separators=self.separators,
            )
        else:
            return RecursiveCharacterTextSplitter.from_tiktoken_encoder(
                encoding_name="",
                model_name=DEFAULT_OPENAI_MODEL
                if not self.model_name
                else self.model_name,
                chunk_size=self.chunk_size,
                chunk_overlap=self.chunk_overlap,
                separators=self.separators,
            )

    def _text_splitter(self) -> RecursiveCharacterTextSplitter:
        return (
            self._characters_splitter()
            if self.split_by_characters
            else self._tokens_splitter()
        )

    def save_document(self, doc: List[Document]) -> str:
        '''
        This function saves document to the vector database
        Args:
            doc List[Document]: document content, read with DocumentReader load_document(), 
                or document_from_pdf(), document_from_txt()
        Returns:
            str: document id, which can be used to remove an uploaded document with remove_document()
        '''
        splitted = self._split_document(doc)
        return self._store_embeddings(splitted)

    def _split_document(self, document: List[Document]) -> List[Document]:
        text_splitter = self._text_splitter()
        return text_splitter.split_documents(document)

    def _store_embeddings(
        self, doc: List[Document]
    ) -> str:
        return self.database_host.store_embeddings(
            documents=doc
        )

    def similarity_search(
        self,
        query: str,
        k: int = 3
    ):
        """
        Returns top n closest matches to query from vector store.

        Args:
            query (str): query string

            k (int, optional): number of closest matches to return. Defaults to
            3.

            collection_name (str): search in collection {collection_name}

        """
        return self.database_host.similarity_search(
            query=query, k=k or self.n_results
        )

    def connect(self, host: str, port: str) -> None:
        self.database_host.connect(host, port)

    def get_all_documents(self) -> List[Dict]:
        return self.database_host.get_all_documents()

    def remove_document(self, doc_id: str) -> None:
        self.database_host.remove_document(doc_id)


class DocumentReader:
    def load_document(self, path: str) -> List[Document]:
        """
        Loads a document from a path; accepts txt and pdf files. Txt files are
        loaded as-is, pdf files are converted to text using fitz.

        Args:
            path (str): path to document

        Returns:
            List[Document]: list of documents
        """
        if path.endswith(".txt"):
            loader = TextLoader(path)
            return loader.load()

        elif path.endswith(".pdf"):
            doc = fitz.open(path)
            text = ""
            for page in doc:
                text += page.get_text()

            meta = {k: v for k, v in doc.metadata.items() if v}
            meta.update({"source": path})

            return [
                Document(
                    page_content=text,
                    metadata=meta,
                )
            ]

    def document_from_pdf(self, pdf: bytes) -> List[Document]:
        """
        Receive a byte representation of a pdf file and return a list of Documents
        with metadata.

        Args:
            pdf (bytes): byte representation of pdf file

        Returns:
            List[Document]: list of documents
        """
        doc = fitz.open(stream=pdf, filetype="pdf")
        text = ""
        for page in doc:
            text += page.get_text()

        meta = {k: v for k, v in doc.metadata.items() if v}
        meta.update({"source": "pdf"})

        return [
            Document(
                page_content=text,
                metadata=meta,
            )
        ]

    def document_from_txt(self, txt: bytes) -> List[Document]:
        """
        Receive a byte representation of a txt file and return a list of Documents
        with metadata.

        Args:
            txt (bytes): byte representation of txt file

        Returns:
            List[Document]: list of documents
        """
        meta = {"source": "txt"}
        return [
            Document(
                page_content=txt,
                metadata=meta,
            )
        ]



from typing import List
from langchain.schema import Document
from .llm_connect import GptConversation
from gtts import gTTS
import nltk
import os

FIRST_PROMPT = (
    "You are tasked with summarising a scientific manuscript for consumption as"
    "a podcast. As a first step, extract title and authors of the manuscript"
    "from the following text. Return them in the format 'Title: <title>,"
    "Authors: <authors>'."
)

PROMPT = (
    "You are tasked with summarising a scientific manuscript for consumption as"
    "a podcast. You will receive a collection of sentences from the"
    "manuscript, from which you will remove any information not relevant to the"
    "content, such as references, figure legends, tables, author information, "
    "journal metadata, and so on. You will then be asked to summarise the"
    "section of the manuscript, making the wording more suitable for listening."
    "Remove all content in brackets that is of technical nature, such as"
    "p-values, statistical tests, and so on. If the given text contains only "
    "literature references, return 'No content'."
)


class Podcaster:
    def __init__(self, document: Document) -> None:
        """
        Orchestrates the podcasting of a document.
        """
        self.document = document

    def generate_podcast(self, characters_per_paragraph: int) -> None:
        """
        Podcasts the document.

        TODO:
        - chain of density prompting for variable summary length
        - do not summarise but just make readable
        """
        full_text = self.document[0].page_content

        # split text by sentence
        sentences = self._split_text(full_text)

        # could embed sentences and cluster on cosine similarity to identify
        # paragraphs here

        # preprocess text
        for i, sentence in enumerate(sentences):
            # special cases i.e. and e.g. - if sentence ends with one of these,
            # append next sentence
            special_cases = ["i.e.", "e.g."]
            if sentence.endswith(tuple(special_cases)):
                sentences[i] = sentence + " " + sentences[i + 1]
                del sentences[i + 1]

        # concatenate first 5 sentences for title and author extraction
        first_5 = "\n".join(sentences[:5])
        self.podcast_intro = self._title_and_authors(first_5)

        # LLM to determine section breaks?

        # go through sections and summarise each
        self.summarised_sections = self._summarise_sections(
            sentences,
            characters_per_paragraph,
        )

        # summarise the summaries

    def _split_text(self, text: str) -> List[str]:
        """
        Splits consecutive text into sentences.
        """
        nltk.download("punkt")
        tokenizer = nltk.data.load("tokenizers/punkt/english.pickle")
        return tokenizer.tokenize(text)

    def _title_and_authors(self, text: str) -> str:
        """
        Extracts title and authors from document.

        Args:
            text (str): text to extract title and authors from

        Returns:
            str: title and authors
        """
        # first sentence - extract title, authors
        c_first = GptConversation(
            model_name="gpt-3.5-turbo",
            prompts={},
            correct=False,
        )
        c_first.set_api_key(api_key=os.getenv("OPENAI_API_KEY"), user="podcast")
        c_first.append_system_message(FIRST_PROMPT)
        msg, token_usage, correction = c_first.query(text)
        # split at authors ('Authors:' or '\nAuthors:')
        title = msg.split("Title:")[1].split("Authors:")[0].strip()
        authors = msg.split("Authors:")[1].strip()
        return f"{title}, by {authors}, podcasted by biochatter."

    def _summarise_section(self, text: str) -> str:
        """
        Summarises a section of the document.

        Args:
            text (str): text to summarise

        Returns:
            str: summarised text
        """
        # summarise section
        c = GptConversation(
            model_name="gpt-3.5-turbo",
            prompts={},
            correct=False,
        )
        c.set_api_key(api_key=os.getenv("OPENAI_API_KEY"), user="podcast")
        c.append_system_message(PROMPT)
        msg, token_usage, correction = c.query(text)
        return msg

    def _summarise_sections(
        self, sentences: list, characters_per_paragraph: int
    ) -> list:
        """

        Summarises sections of the document. Concatenates sentences until
        characters_per_paragraph is reached, removing each sentence from the
        list as it is added to the section to be summarised.

        Args:
            sentences (list): list of sentences to summarise
            characters_per_paragraph (int): number of characters per paragraph

        Returns:
            list: list of summarised sections
        """
        summarised_sections = []
        section = ""
        while sentences:
            sentence = sentences.pop(0)
            tmp = section + sentence
            if len(tmp) < characters_per_paragraph and sentences:
                section += sentence
            else:
                if sentences:
                    sentences.insert(0, sentence)
                summarised_section = self._summarise_section(section)
                summarised_sections.append(summarised_section)
                section = ""

        return summarised_sections

    def podcast_to_file(self, path: str) -> None:
        """
        Uses text-to-speech to generate audio for the summarised paper podcast.

        Args:
            path (str): path to save audio file to
        """

        full_text = self.podcast_to_text()

        audio = gTTS(text=full_text)
        audio.save(path)

    def podcast_to_text(self):
        """
        Returns the summarised paper podcast as text.
        """
        full_text = self.podcast_intro + "\n\n"
        for section in self.summarised_sections:
            full_text += section + "\n\n"
        return full_text



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['biochatter']

package_data = \
{'': ['*']}

install_requires = \
['langchain==0.0.147',
 'nltk>=3.8.1,<4.0.0',
 'openai>=0.27.8,<0.28.0',
 'pymilvus==2.2.8',
 'pymupdf>=1.22.3,<2.0.0',
 'redis>=4.5.5,<5.0.0',
 'retry>=0.9.2,<0.10.0',
 'stringcase>=1.2.0,<2.0.0',
 'tiktoken>=0.4.0,<0.5.0',
 'transformers>=4.30.2,<5.0.0']

extras_require = \
{'podcast': ['gTTS>=2.3.2,<3.0.0'], 'streamlit': ['streamlit>=1.23.1,<2.0.0']}

setup_kwargs = {
    'name': 'biochatter',
    'version': '0.1.15',
    'description': 'Backend library for conversational AI in biomedicine',
    'long_description': '# biochatter\n\nThis repository contains the `biochatter` Python package, a generic backend\nlibrary for the connection of biomedical applications to conversational AI.\nDescribed in [this preprint](https://arxiv.org/abs/2305.06488) and used in\n[ChatGSE](https://chat.biocypher.org), which is being developed at\nhttps://github.com/biocypher/ChatGSE. More to come, so stay tuned!\n\n## Installation\n\nTo use the package, install it from PyPI, for instance using pip (`pip install\nbiochatter`) or Poetry (`poetry add biochatter`).\n\n### Extras\n\nThe package has some optional dependencies that can be installed using the\nfollowing extras (e.g. `pip install biochatter[streamlit]`):\n\n- `streamlit`: support for streamlit UI functions (used in ChatGSE)\n- `podcast`: support for podcast text-to-speech\n\n## Usage\n\nAs an interim documentation until we have a proper one, check out the\n[Wiki](https://github.com/biocypher/biochatter/wiki) for some usage examples.\nMany common functionalities covered by BioChatter can be seen in use in the\n[ChatGSE](https://github.com/biocypher/ChatGSE) code base.\n\n# More information about LLMs\n\nCheck out [this repository](https://github.com/csbl-br/awesome-compbio-chatgpt)\nfor more info on computational biology usage of large language models.\n\n# Dev Container\n\nDue to some incompatibilities of `pymilvus` with Apple Silicon, we have created\na dev container for this project. To use it, you need to have Docker installed\non your machine. Then, you can run the devcontainer setup as recommended by\nVSCode\n[here](https://code.visualstudio.com/docs/remote/containers#_quick-start-open-an-existing-folder-in-a-container)\nor using Docker directly.\n\nThe dev container expects an environment file (there are options, but the basic\none is `.devcontainer/local.env`) with the following variables:\n\n```\nOPENAI_API_KEY=(sk-...)\nDOCKER_COMPOSE=true\nDEVCONTAINER=true\n```\n\nTo test vector database functionality, you also need to start a Milvus\nstandalone server. You can do this by running `docker-compose up` as described\n[here](https://milvus.io/docs/install_standalone-docker.md) on the host machine\n(not from inside the devcontainer).\n',
    'author': 'Sebastian Lobentanzer',
    'author_email': 'sebastian.lobentanzer@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.10,<3.12',
}


setup(**setup_kwargs)
